#!/usr/bin/env python
# -*- coding: utf-8 -*-
##********************************************************************************************************************************************************
##
##  This module converts molfit file into MAGIX instance xml file
##  Copyright (C) 2009 - 2016  Thomas Moeller
##
##  I. Physikalisches Institut, University of Cologne
##
##
##
##  The following subroutines and functions are included in this module:
##
##      - subroutine ConversionMolfit2XML:              main subroutine
##
##
##
##  Versions of the program:
##
##  Who           When         What
##
##  T. Moeller    22.06.2009   initial version
##  T. Moeller    16.01.2012   improve documentation of source code
##  T. Moeller    08.01.2013   include new molfit file formats
##
##
##
##  License:
##
##    GNU GENERAL PUBLIC LICENSE
##    Version 3, 29 June 2007
##    (Copyright (C) 2007 Free Software Foundation, Inc. <http://fsf.org/>)
##
##
##    This program is free software: you can redistribute it and/or modify
##    it under the terms of the GNU General Public License as published by
##    the Free Software Foundation, either version 3 of the License, or
##    (at your option) any later version.
##
##    This program is distributed in the hope that it will be useful,
##    but WITHOUT ANY WARRANTY; without even the implied warranty of
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##    GNU General Public License for more details.
##
##    You should have received a copy of the GNU General Public License
##    along with this program.  If not, see <http://www.gnu.org/licenses/>.
##
##********************************************************************************************************************************************************


"""
Program molfit2xml:
-------------------

This program converts a file in the molfit format to a xml-file structure used for the MAGIX program
package. Note, that the molfit file has to end with ".molfit" and the xml-file name has to end with ".xml"

"""

import numpy                                                                                ## import numpy package
import os.path                                                                              ## load package for file management
import math                                                                                 ## import math package
import sys                                                                                  ## load package sys


##********************************************************************************************************************************************************
def ConversionMolfit2XML(print_flag, molfitFilename, xmlFileName):
    ## read registration file
    """

    ConversionMolfit2XML: This subroutine converts a molfit file used by myXCLASS to a xml-file used by MAGIX.

    input parameter:    print_flag:                 flag if screen output is permitted (=true) or not
                        molfitFilename:             path and file name of the molfit file
                        xmlFileName:                path and file name of the xml description of the model

    output parameter:   ok:                         status; if everything is ok: ok = 0

    """

    ok = 0


    ##----------------------------------------------------------------------------------------------------------------------------------------------------
    ## get molfit and xml filename
    if not molfitFilename.endswith(".molfit"):
        print " "
        print "\t Error in molfit2xml subroutine!"
        print "\t     The selected file ",molfitFilename
        print "\t     is not a molfit file!"
        print "\t     Please correct input!"
        print " "
        ok = 1
        return

    if not xmlFileName.endswith(".xml"):
        print " "
        print "\t Error in molfit2xml subroutine!"
        print "\t     The selected file ",xmlFileName
        print "\t     is not a xml-file!"
        print "\t     Please correct input!"
        print " "
        ok = 1
        return


    # Debug:
    # print 'molfitFilename = ',molfitFilename
    # print 'xmlFileName = ',xmlFileName


    ##----------------------------------------------------------------------------------------------------------------------------------------------------
    ## read molfit file
    NumberParameter = 0
    NumberMolecules = 0
    ParameterList = []
    molfitFile = open(molfitFilename)                                                       ## open molfit file
    try:
        for line in molfitFile:                                                             ## loop over all lines in the molfit File


            ## test, if line is remarked
            reducedLine = line.strip(" ")                                                   ## remove leading blanks
            if not reducedLine.startswith("%"):                                             ## starts line without "%" character ?


                ## remove comments
                i = line.find("%")                                                          ## is there a comment in the current line
                if (i > 0):
                    pureline = line[:i]                                                     ## remove comment
                else:
                    pureline = line


                ## split lines in parts
                ElementsOfLine = pureline.split()                                           ## determine elements of the current line
                NumberElementsLine = len(ElementsOfLine)                                    ## determine number of elements of the current line
                if (NumberElementsLine > 1):                                                ## ignore current line if there is less than one column


                    ## does the current line define a name of a molecule?
                    LastElement = ElementsOfLine[-1]                                        ## get the last element
                    NameLineFlag = "false"
                    if (LastElement.replace('.','a',1).isdigit()):                          ## is last element an integer number (here, we convert
                        NameLineFlag = "true"                                               ## an floating point number to a string)

                    # Debug:
                    # print "NameLineFlag = ", NameLineFlag


                    ## if current line defines name of molecule
                    if (NameLineFlag == "true"):
                        NumberParameter += 2
                        NumberMolecules += 1                                                ## count number of molecules
                        NameMolecule = ""
                        counter = 0
                        for element in ElementsOfLine:
                            counter += 1
                            if (counter < NumberElementsLine):                              ## don't include the number of components into the name
                                NameMolecule += " " + element
                        ParameterList.append(["Molecule_Name", 0, 0, 0, NameMolecule])
                        ParameterList.append(["Number_MolLine", 0, 0, 0, LastElement])

                        # Debug:
                        # print "NameMolecule = ", NameMolecule
                        # print "LastElement = ", LastElement


                    ## continue here if the current line does not define a component
                    else:
                        # Debug:
                        # print "NumberElementsLine = ", NumberElementsLine
                        # print "pureline = ", pureline.strip()


                        LastElement = LastElement.lower()
                        if (LastElement != "a" and LastElement != "e"):
                            pureline = pureline + " e"
                            ElementsOfLine = pureline.split()                               ## determine elements of the current line
                            NumberElementsLine = len(ElementsOfLine)                        ## determine number of elements of the current line


                        ## determine version of myXCLASS
                        myXCLASSVersion = ""
                        lowline = pureline.lower()                                          ## convert line to lower case
                        tfFlag = "no"
                        if (lowline.find("y") > (-1) or lowline.find("n") > (-1) or lowline.find("t") > (-1) or lowline.find("f") > (-1)):
                            tfFlag = "yes"

                        # Debug:
                        # print "lowline = ", lowline.strip()
                        # print "NumberElementsLine = ", NumberElementsLine


                        ## old version
                        if (NumberElementsLine == 11):
                            myXCLASSVersion = "OLDtbFnhF"
                        elif (NumberElementsLine == 15):
                            myXCLASSVersion = "OLDtbTnhF"
                        elif (NumberElementsLine == 17):
                            myXCLASSVersion = "OLDtbFnhT"
                        elif (NumberElementsLine == 21 and tfFlag == "no"):
                            myXCLASSVersion = "OLDtbTnhT"


                        ## new version
                        elif (NumberElementsLine == 21 and tfFlag == "yes"):
                            myXCLASSVersion = "NEWtbFnhF"
                        elif (NumberElementsLine == 29):
                            myXCLASSVersion = "NEWtbTnhF"
                        elif (NumberElementsLine == 33):
                            myXCLASSVersion = "NEWtbFnhT"
                        elif (NumberElementsLine == 41):
                            myXCLASSVersion = "NEWtbTnhT"
                        else:
                            print ""
                            print "Error in subroutine ConversionMolfit2XML:"
                            print "  Can not identify the format of the current line"
                            print "  in molfit file " + molfitFilename
                            print " "

                        # Debug:
                        # print "tfFlag = ", tfFlag
                        # print "myXCLASSVersion = ", myXCLASSVersion


                        ## analyse current line
                        flag = 0
                        uplimit = 0
                        lowlimit = 0
                        counter = 0
                        for element in ElementsOfLine:                                      ## loop over all elements of the current line
                            counter += 1                                                    ## increase counter for elements in the current line
                            NumberParameter += 1


                            ## analyse current line if format of molfit file is "old"
                            if (myXCLASSVersion[:3] == "OLD"):
                                if (counter == 1 or counter == 3 or counter == 5 or counter == 7 or counter == 9):
                                    flag = element.strip()
                                    Name = ""
                                elif (counter == 2):                                        ## get value for source_size
                                    Name = "source_size"
                                elif (counter == 4):                                        ## get value for T_rot
                                    Name = "T_rot"
                                elif (counter == 6):                                        ## get value for N_tot
                                    Name = "N_tot"
                                elif (counter == 8):                                        ## get value for V_width
                                    Name = "V_width"
                                elif (counter == 10):                                       ## get value for V_off
                                    Name = "V_off"


                                ## distinguish between different formats
                                if (myXCLASSVersion == "OLDtbFnhF"):
                                    if (counter == 11):                                     ## get value for AbsorptionFlag
                                        Name = "AbsorptionFlag"
                                        element = "  " + element
                                        flag = 0

                                elif (myXCLASSVersion == "OLDtbTnhF"):
                                    if (counter == 11 or counter == 13):
                                        flag = element.strip()
                                        Name = ""
                                    elif (counter == 12):                                   ## get value for T_Back
                                        Name = "T_Back"
                                    elif (counter == 14):                                   ## get value for T_Slope
                                        Name = "T_Slope"
                                    elif (counter == 15):                                   ## get value for AbsorptionFlag
                                        Name = "AbsorptionFlag"
                                        element = "  " + element
                                        flag = 0

                                elif (myXCLASSVersion == "OLDtbFnhT"):
                                    if (counter == 11 or counter == 13 or counter == 15):
                                        flag = float(element.strip())
                                        Name = ""
                                    elif (counter == 12):                                   ## get value for nHcolumn
                                        Name = "nHcolumn"
                                    elif (counter == 14):                                   ## get value for kappa
                                        Name = "kappa"
                                    elif (counter == 16):                                   ## get value for beta
                                        Name = "beta"
                                    elif (counter == 17):                                   ## get value for AbsorptionFlag
                                        Name = "AbsorptionFlag"
                                        element = "  " + element
                                        flag = 0

                                elif (myXCLASSVersion == "OLDtbTnhT"):
                                    if (counter == 11 or counter == 13 or counter == 15 or counter == 17 or counter == 19):
                                        flag = element.strip()
                                        Name = ""
                                    elif (counter == 12):                                   ## get value for T_Back
                                        Name = "T_Back"
                                    elif (counter == 14):                                   ## get value for T_Slope
                                        Name = "T_Slope"
                                    elif (counter == 16):                                   ## get value for nHcolumn
                                        Name = "nHcolumn"
                                    elif (counter == 18):                                   ## get value for kappa
                                        Name = "kappa"
                                    elif (counter == 20):                                   ## get value for beta
                                        Name = "beta"
                                    elif (counter == 21):                                   ## get value for AbsorptionFlag
                                        Name = "AbsorptionFlag"
                                        element = "  " + element
                                        flag = 0


                                ## calculate upper and lower limits and create new entry for ParameterList array
                                if (Name != ""):
                                    if (Name != "AbsorptionFlag"):
                                        entry = [Name + "_FitFlag", 0, 0, 0, flag]
                                        ParameterList.append(entry)


                                    ## calculate upper and lower limits
                                    if (float(flag) == 0):
                                        uplimit = 0
                                        lowlimit = 0
                                    else:
                                        ParameterValue = (float(element))
                                        LimitValues = abs(float(flag))
                                        if (Name == "source_size" or Name == "N_tot"):
                                            lowlimit = ParameterValue / LimitValues         ## modification by Tom Bell
                                            if (lowlimit < 0):
                                                lowlimit = 0
                                            uplimit = ParameterValue * LimitValues          ## modification by Tom Bell
                                        else:
                                            lowlimit = ParameterValue - LimitValues
                                            if (lowlimit < 0 and Name != "V_off" and Name != "T_Back"): ## V_off, T_Back could be negativ !!!
                                                lowlimit = 0
                                            uplimit = ParameterValue + LimitValues

                                    entry = [Name, flag, lowlimit, uplimit, element]
                                    ParameterList.append(entry)


                            ## analyse current line if format of molfit file is "new"
                            else:
                                if (counter == 1 or counter == 5 or counter == 9 or counter == 13 or counter == 17):
                                    flag = element.strip()                                  ## get flag for fitting
                                    Name = ""
                                elif (counter == 2 or counter == 6 or counter == 10 or counter == 14 or counter == 18):
                                    uplimit = float(element.strip())                        ## get upper limit
                                    Name = ""
                                elif (counter == 3 or counter == 7 or counter == 11 or counter == 15 or counter == 19):
                                    lowlimit = float(element.strip())                       ## get lower limit
                                    Name = ""
                                elif (counter == 4):                                        ## get value for source_size
                                    Name = "source_size"
                                elif (counter == 8):                                        ## get value for T_rot
                                    Name = "T_rot"
                                elif (counter == 12):                                       ## get value for N_tot
                                    Name = "N_tot"
                                elif (counter == 16):                                       ## get value for V_width
                                    Name = "V_width"
                                elif (counter == 20):                                       ## get value for V_off
                                    Name = "V_off"


                                ## distinguish between different formats
                                if (myXCLASSVersion == "NEWtbFnhF"):
                                    if (counter == 21):                                     ## get value for AbsorptionFlag
                                        Name = "AbsorptionFlag"
                                        element = "  " + element
                                        flag = "0"
                                        uplimit = 0
                                        lowlimit = 0

                                elif (myXCLASSVersion == "NEWtbTnhF"):
                                    if (counter == 21 or counter == 25):
                                        flag = element.strip()                              ## get flag for fitting
                                        Name = ""
                                    elif (counter == 22 or counter == 26):
                                        uplimit = float(element.strip())                    ## get upper limit
                                        Name = ""
                                    elif (counter == 23 or counter == 27):
                                        lowlimit = float(element.strip())                   ## get lower limit
                                        Name = ""
                                    elif (counter == 24):                                   ## get value for T_Back
                                        Name = "T_Back"
                                    elif (counter == 28):                                   ## get value for T_Slope
                                        Name = "T_Slope"
                                    elif (counter == 29):                                   ## get value for AbsorptionFlag
                                        Name = "AbsorptionFlag"
                                        element = "  " + element
                                        flag = "0"
                                        uplimit = 0
                                        lowlimit = 0


                                elif (myXCLASSVersion == "NEWtbFnhT"):
                                    if (counter == 21 or counter == 25 or counter == 29):
                                        flag = element.strip()                              ## get flag for fitting
                                        Name = ""
                                    elif (counter == 22 or counter == 26 or counter == 30):
                                        uplimit = float(element.strip())                    ## get upper limit
                                        Name = ""
                                    elif (counter == 23 or counter == 27 or counter == 31):
                                        lowlimit = float(element.strip())                   ## get lower limit
                                        Name = ""
                                    elif (counter == 24):                                   ## get value for nHcolumn
                                        Name = "nHcolumn"
                                    elif (counter == 28):                                   ## get value for kappa
                                        Name = "kappa"
                                    elif (counter == 32):                                   ## get value for beta
                                        Name = "beta"
                                    elif (counter == 33):                                   ## get value for AbsorptionFlag
                                        Name = "AbsorptionFlag"
                                        element = "  " + element
                                        flag = "0"
                                        uplimit = 0
                                        lowlimit = 0

                                elif (myXCLASSVersion == "NEWtbTnhT"):
                                    if (counter == 21 or counter == 25 or counter == 29 or counter == 33 or counter == 37):
                                        flag = element.strip()                              ## get flag for fitting
                                        Name = ""
                                    elif (counter == 22 or counter == 26 or counter == 30 or counter == 34 or counter == 38):
                                        uplimit = float(element.strip())                    ## get upper limit
                                        Name = ""
                                    elif (counter == 23 or counter == 27 or counter == 31 or counter == 35 or counter == 39):
                                        lowlimit = float(element.strip())                   ## get lower limit
                                        Name = ""
                                    elif (counter == 24):                                   ## get value for T_Back
                                        Name = "T_Back"
                                    elif (counter == 28):                                   ## get value for T_Slope
                                        Name = "T_Slope"
                                    elif (counter == 32):                                   ## get value for nHcolumn
                                        Name = "nHcolumn"
                                    elif (counter == 36):                                   ## get value for kappa
                                        Name = "kappa"
                                    elif (counter == 40):                                   ## get value for beta
                                        Name = "beta"
                                    elif (counter == 41):                                   ## get value for AbsorptionFlag
                                        Name = "AbsorptionFlag"
                                        element = "  " + element
                                        flag = "0"
                                        uplimit = 0
                                        lowlimit = 0


                                ## create new entry for ParameterList array
                                if (Name != ""):
                                    if (Name != "AbsorptionFlag"):
                                        entry = [Name + "_flag", 0, 0, 0, flag]
                                        ParameterList.append(entry)
                                        entry = [Name + "_uplimit", 0, 0, 0, uplimit]
                                        ParameterList.append(entry)
                                        entry = [Name + "_lowlimit", 0, 0, 0, lowlimit]
                                        ParameterList.append(entry)

                                    if (flag.find("y") > (-1) or flag.find("t") > (-1)):
                                        flag = 1
                                    else:
                                        flag = 0
                                    entry = [Name, flag, lowlimit, uplimit, element]
                                    ParameterList.append(entry)

            # Debug:
            # else:
            #     print ">",line.strip(),"<"

    finally:                                                                                ## close molfit file
        molfitFile.close()


    ##----------------------------------------------------------------------------------------------------------------------------------------------------
    ## write parameter.xml file
    file = open(xmlFileName, 'w')
    line="<?xml version=" + chr(34) + "1.0" + chr(34) + " ?>\n"
    file.write(line)
    file.write("<ModelParameters>\n")


    ## write number of parameters
    file.write("\n")
    file.write("\n")
    file.write("    <!-- define number of parameters -->\n")
    file.write("    <NumberParameters>" + str(NumberParameter + 4) + "</NumberParameters>\n")
    file.write("\n")
    file.write("\n")
    file.write("    <!-- ############################################################################################################################################# -->\n")
    file.write("    <!-- first input file -->\n")
    file.write("\n")
    file.write("\n")


    ## write number of molecules to xml-file
    file.write("    <!-- define number of molecules -->\n")
    file.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
    file.write("        <name>Number_Molecules</name>\n")
    file.write("        <value>" + str(NumberMolecules) + "</value>\n")
    file.write("        <error> </error>\n")
    file.write("        <lowlimit>0</lowlimit>\n")
    file.write("        <uplimit>1</uplimit>\n")
    file.write("    </Parameter>\n")
    file.write("\n")
    file.write("\n")


    ## write parameter indicating molfit format
    file.write("    <!-- define myXCLASSVersion -->\n")
    file.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
    file.write("        <name>myXCLASSVersion</name>\n")
    file.write("        <value>" + myXCLASSVersion[:3] + "</value>\n")
    file.write("        <error> </error>\n")
    file.write("        <lowlimit>0</lowlimit>\n")
    file.write("        <uplimit>1</uplimit>\n")
    file.write("    </Parameter>\n")
    file.write("\n")
    file.write("\n")


    ## write parameter tbFlag indicating if background temperature etc is given in molfit file
    file.write("    <!-- define tbFlag -->\n")
    file.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
    file.write("        <name>tbFlag</name>\n")
    file.write("        <value>" + myXCLASSVersion[5:6] + "</value>\n")
    file.write("        <error> </error>\n")
    file.write("        <lowlimit>0</lowlimit>\n")
    file.write("        <uplimit>1</uplimit>\n")
    file.write("    </Parameter>\n")
    file.write("\n")
    file.write("\n")


    ## write parameter nHFlag indicating if column density etc is given in molfit file
    file.write("    <!-- define nHFlag -->\n")
    file.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
    file.write("        <name>nHFlag</name>\n")
    file.write("        <value>" + myXCLASSVersion[8:9] + "</value>\n")
    file.write("        <error> </error>\n")
    file.write("        <lowlimit>0</lowlimit>\n")
    file.write("        <uplimit>1</uplimit>\n")
    file.write("    </Parameter>\n")
    file.write("\n")
    file.write("\n")


    ## write parameter to xml-file
    for param in ParameterList:                                                             ## loop over all parameters
        ParameterName = param[0].strip()                                                    ## get name of parameter
        LimitValues = param[1]                                                              ## get fit-flag of parameter
        LimitValuesTrueFalse = "false"                                                      ## define LimitValuesTrueFalse variable
        if (float(LimitValues) != 0):
            LimitValuesTrueFalse = "true"
        lowlimit = param[2]                                                                 ## get lower limit
        uplimit = param[3]                                                                  ## get upper limit
        ParameterValue = param[4]                                                           ## get value of parameter

        # Debug:
        # print "param = ", param


        ##------------------------------------------------------------------------------------------------------------------------------------------------
        ## for column and hydrogen column density use a log scale for fit
        if (ParameterName.strip() == "N_tot" or (ParameterName.strip() == "nHcolumn" and myXCLASSVersion[-3:].lower() == "nht")):
            ParameterValue = float(ParameterValue)

            # Debug:
            # print "ParameterValue = ", ParameterValue
            # print "lowlimit = ", lowlimit
            # print "uplimit = ", uplimit


            if (ParameterValue != 0.E0):
                if (ParameterValue < 0.E0):
                    ParameterValue = abs(ParameterValue)
                ParameterValue = math.log10(ParameterValue)                                 ## convert value to log scale
                if (lowlimit != 0.E0 or uplimit != 0.E0):                                   ## continue here, only if limit unequal zero


                    ## check and convert lower limit
                    if (lowlimit == 0.E0):                                                  ## is lower limit exactly zero?
                        lowlimit = 1.e-99                                                   ## set to a very small number
                    elif (lowlimit < 0.E0):                                                 ## is lower limit less than zero?
                        lowlimit = abs(lowlimit)                                            ## take the absolute value
                    lowlimit = math.log10(lowlimit)                                         ## convert value to log scale
                    if (lowlimit > 307.0):                                                  ## check for overflow
                        lowlimit = 307.0

                    # Debug:
                    # print ">>lowlimit = ", lowlimit


                    ## check and convert upper limit
                    if (uplimit == 0.E0):                                                   ## is upper limit exactly zero?
                        uplimit = 1.e-99                                                    ## set to a very small number
                    elif (uplimit < 0.E0):                                                  ## is upper limit less than zero?
                        uplimit = abs(uplimit)                                              ## take the absolute value
                    uplimit = math.log10(uplimit)                                           ## convert value to log scale
                    if (uplimit > 307.0):                                                   ## check for overflow
                        uplimit = 307.0

                    # Debug:
                    # print ">>uplimit = ", uplimit
        ##------------------------------------------------------------------------------------------------------------------------------------------------


        ## for column and hydrogen column density use a log scale for fit
        if (myXCLASSVersion[:3] == "NEW"):
            if (ParameterName.strip() == "N_tot_uplimit" or ParameterName.strip() == "N_tot_lowlimit"):
                ParameterValue = float(ParameterValue)
                if (ParameterValue != 0.E0):
                    if (ParameterValue < 0.E0):
                        ParameterValue = abs(ParameterValue)
                    ParameterValue = math.log10(ParameterValue)                             ## convert value to log scale

            if (myXCLASSVersion[-3:].lower() == "nht"):
                if (ParameterName.strip() == "nHcolumn_uplimit" or ParameterName.strip() == "nHcolumn_lowlimit"):
                    ParameterValue = float(ParameterValue)
                    if (ParameterValue != 0.E0):
                        if (ParameterValue < 0.E0):
                            ParameterValue = abs(ParameterValue)
                        ParameterValue = math.log10(ParameterValue)                         ## convert value to log scale


        ## write name, value and optimization status for each parameter to xml file
        file.write("    <!-- define " + ParameterName + " -->\n")
        file.write("    <Parameter fit=" + chr(34) + LimitValuesTrueFalse + chr(34) + ">\n")
        file.write("        <name>" + ParameterName + "</name>\n")
        file.write("        <value>" + str(ParameterValue).strip() + "</value>\n")
        file.write("        <error> </error>\n")


        ## check order of limit values
        if (lowlimit > uplimit):                                                            ## check, if upper and lower limit values have still
            h = uplimit                                                                     ## correct order
            uplimit = lowlimit
            lowlimit = h


        ## write limits for each parameter to xml-file
        if (LimitValuesTrueFalse=="false"):
            file.write("        <lowlimit>-1e9</lowlimit>\n")
            file.write("        <uplimit>1e9</uplimit>\n")
        else:
            file.write("        <lowlimit>" + str(lowlimit) + "</lowlimit>\n")
            file.write("        <uplimit>" + str(uplimit) + "</uplimit>\n")
        file.write("    </Parameter>\n")
        file.write("\n")
        file.write("\n")


    ## write end of xml-file
    file.write("</ModelParameters>\n")


    ## close xml-file
    file.close()

    return ok
##--------------------------------------------------------------------------------------------------------------------------------------------------------


##********************************************************************************************************************************************************
def GetIsoTableParameters(printflag, IsoTableFileName, parameter_file):
    ## read iso table parameters from iso table file
    """

    GetIsoTableParameters: This subroutine reads in parameters from the iso table file

    input parameter:    printflag:                  flag if screen output is permitted (=true) or not
                        IsoTableFileName:           path and file name of the iso file
                        parameter_file:             path and file name of the instance XML file

    output parameter:   ok:                         status; if everything is ok: ok = 0

    """

    ok = 0                                                                                  ## reset status variable


    ## read in contents of iso file
    counterIsoRatios = 0
    if (IsoTableFileName.strip() != ""):
        IsoFile = open(IsoTableFileName)
        IsoFileContents = IsoFile.readlines()
        IsoFile. close()


        ## determine number of iso ratios in iso table file and write number of iso ratios to xml-file
        for line in IsoFileContents:
            StrippedLine = line.strip()                                                     ## comment lines starts with "%"
            i = StrippedLine.find("%")
            if (i > (-1)):
                StrippedLine = StrippedLine[:i].strip()
            if (StrippedLine != ""):                                                        ## ignore empty lines
                counterIsoRatios += 1
    #else:
    #    return ok


    ## read in contents of instance XML file
    ParameterFile = open(parameter_file)
    ParameterFileContents = ParameterFile.readlines()
    ParameterFile. close()


    ## write first part of instance XML file back to instance file
    NewParameterFile = open(parameter_file, 'w')
    counterLine = 0
    for line in ParameterFileContents:
        counterLine += 1
        if (counterLine == 6):
            Newline = line.replace("<NumberParameters>", " ")
            Newline = Newline.replace("</NumberParameters>", " ")
            NumParam = int(Newline.strip())

            # Debug:
            # print "Newline = >>", Newline.strip(), "<<"
            # print "NumParam = ", NumParam

            NumParam += (counterIsoRatios * 5) + 1
            NewParameterFile.write("    <NumberParameters>" + str(NumParam) + "</NumberParameters>\n")
        else:
            if (counterLine < len(ParameterFileContents)):
                NewParameterFile.write(line)


    ## write separator
    NewParameterFile.write("\n")
    NewParameterFile.write("\n")
    NewParameterFile.write("    <!-- ############################################################################################################################################# -->\n")
    NewParameterFile.write("    <!-- third input file -->\n")
    NewParameterFile.write("\n")
    NewParameterFile.write("\n")
    NewParameterFile.write("    <!-- define number of iso ratios -->\n")
    NewParameterFile.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
    NewParameterFile.write("        <name>NumberOfISOLines</name>\n")
    NewParameterFile.write("        <value>" + str(counterIsoRatios) + "</value>\n")
    NewParameterFile.write("        <error> </error>\n")
    NewParameterFile.write("        <lowlimit>0</lowlimit>\n")
    NewParameterFile.write("        <uplimit>10000</uplimit>\n")
    NewParameterFile.write("    </Parameter>\n")


    ## analyze iso file
    if (counterIsoRatios > 0):
        counterLines = 0
        for line in IsoFileContents:
            StrippedLine = line.strip()                                                     ## comment lines starts with "%"
            i = StrippedLine.find("%")
            if (i > (-1)):
                StrippedLine = StrippedLine[:i].strip()
            if (StrippedLine != ""):                                                        ## ignore empty lines
                SplittedLine = StrippedLine.split()                                         ## cut line in columns
                counterLines += 1                                                           ## increase counter for lines in iso file


                ## write comment to xml file
                NewParameterFile.write("\n")
                NewParameterFile.write("\n")
                NewParameterFile.write("    <!-- line " + str(counterLines) + ":  ratio: " + SplittedLine[0] + " / " + SplittedLine[1] + " -->\n")


                ## write name of iso molecule
                NewParameterFile.write("\n")
                NewParameterFile.write("\n")
                NewParameterFile.write("    <!-- define IsoMolecule -->\n")
                NewParameterFile.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
                NewParameterFile.write("        <name>IsoMolecule</name>\n")
                NewParameterFile.write("        <value>" + SplittedLine[0].strip() + "</value>\n")
                NewParameterFile.write("        <error> </error>\n")
                NewParameterFile.write("        <lowlimit>-1e9</lowlimit>\n")
                NewParameterFile.write("        <uplimit>1e9</uplimit>\n")
                NewParameterFile.write("    </Parameter>\n")


                ## write name of iso molecule
                NewParameterFile.write("\n")
                NewParameterFile.write("\n")
                NewParameterFile.write("    <!-- define IsoMasters -->\n")
                NewParameterFile.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
                NewParameterFile.write("        <name>IsoMasters</name>\n")
                NewParameterFile.write("        <value>" + SplittedLine[1].strip() + "</value>\n")
                NewParameterFile.write("        <error> </error>\n")
                NewParameterFile.write("        <lowlimit>-1e9</lowlimit>\n")
                NewParameterFile.write("        <uplimit>1e9</uplimit>\n")
                NewParameterFile.write("    </Parameter>\n")


                ## check limit specification
                fitFlag = "false"
                if (len(SplittedLine) == 5):
                    lowerlimit = float(SplittedLine[3].strip())
                    upperlimit = float(SplittedLine[4].strip())
                    if (lowerlimit < upperlimit):
                        fitFlag = "true"


                ## write name of iso ratio
                NewParameterFile.write("\n")
                NewParameterFile.write("\n")
                NewParameterFile.write("    <!-- define IsoMasters -->\n")
                if (fitFlag == "true"):
                    NewParameterFile.write("    <Parameter fit=" + chr(34) + "true" + chr(34) + ">\n")
                else:
                    NewParameterFile.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
                NewParameterFile.write("        <name>IsoRatio</name>\n")
                NewParameterFile.write("        <value>" + SplittedLine[2].strip() + "</value>\n")
                NewParameterFile.write("        <error> </error>\n")
                if (fitFlag == "true"):
                    NewParameterFile.write("        <lowlimit>" + SplittedLine[3].strip() + "</lowlimit>\n")
                    NewParameterFile.write("        <uplimit>" + SplittedLine[4].strip() + "</uplimit>\n")
                else:
                    NewParameterFile.write("        <lowlimit>-1e9</lowlimit>\n")
                    NewParameterFile.write("        <uplimit>1e9</uplimit>\n")
                NewParameterFile.write("    </Parameter>\n")


                ## write value of lower limit
                NewParameterFile.write("\n")
                NewParameterFile.write("\n")
                NewParameterFile.write("    <!-- define IsoLowerLimit -->\n")
                NewParameterFile.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
                NewParameterFile.write("        <name>IsoLowerLimit</name>\n")
                if (fitFlag == "true"):
                    NewParameterFile.write("        <value>" + SplittedLine[3].strip() + "</value>\n")
                else:
                    NewParameterFile.write("        <value>-1e9</value>\n")
                NewParameterFile.write("        <error> </error>\n")
                NewParameterFile.write("        <lowlimit>-1e9</lowlimit>\n")
                NewParameterFile.write("        <uplimit>1e9</uplimit>\n")
                NewParameterFile.write("    </Parameter>\n")


                ## write value of lower limit
                NewParameterFile.write("\n")
                NewParameterFile.write("\n")
                NewParameterFile.write("    <!-- define IsoUpperLimit -->\n")
                NewParameterFile.write("    <Parameter fit=" + chr(34) + "false" + chr(34) + ">\n")
                NewParameterFile.write("        <name>IsoUpperLimit</name>\n")
                if (fitFlag == "true"):
                    NewParameterFile.write("        <value>" + SplittedLine[4].strip() + "</value>\n")
                else:
                    NewParameterFile.write("        <value>-1e9</value>\n")
                NewParameterFile.write("        <error> </error>\n")
                NewParameterFile.write("        <lowlimit>-1e9</lowlimit>\n")
                NewParameterFile.write("        <uplimit>1e9</uplimit>\n")
                NewParameterFile.write("    </Parameter>\n")


    ## write end of xml-file
    NewParameterFile.write("</ModelParameters>\n")


    ## close new parameter file
    NewParameterFile.close()


    ## return to MAGIX
    return ok

